# GPUtil - GPU utilization
#
# A Python module for programmically getting the GPU
# utilization from NVIDA GPUs using nvidia-smi
#
# Author: Anders Krogh Mortensen (anderskm)
# Date:   16 January 2017
# Web:    https://github.com/anderskm/gputil
#
# LICENSE
#
# MIT License
#
# Copyright (c) 2017 anderskm
#
# Permission is hereby granted, free of charge, to any
# person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal
# in the Software without restriction, including without
# limitation the rights # to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission
# notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from subprocess import Popen, PIPE
import shutil
import os
import math
import random
import time
import sys
import platform

__version__ = "1.4.1"


class GPU:
    def __init__(
            self,
            ID,
            uuid,
            load,
            memoryTotal,
            memoryUsed,
            memoryFree,
            driver,
            gpu_name,
            serial,
            display_mode,
            display_active,
            temp_gpu,
    ):
        self.id = ID
        self.uuid = uuid
        self.load = load
        self.memoryUtil = float(memoryUsed) / float(memoryTotal)
        self.memoryTotal = memoryTotal
        self.memoryUsed = memoryUsed
        self.memoryFree = memoryFree
        self.driver = driver
        self.name = gpu_name
        self.serial = serial
        self.display_mode = display_mode
        self.display_active = display_active
        self.temperature = temp_gpu


def safeFloatCast(strNumber):
    try:
        number = float(strNumber)
    except ValueError:
        number = float("nan")
    return number


def getGPUs():
    if platform.system() == "Windows":
        # If the platform is Windows and nvidia-smi
        # could not be found from the environment path,
        # try to find it from system drive with default installation path
        nvidia_smi = shutil.which("nvidia-smi")
        if nvidia_smi is None:
            nvidia_smi = (
                "%s\\Program Files\\NVIDIA "
                "Corporation\\NVSMI\\nvidia-smi.exe"
                % os.environ["systemdrive"]
            )
    else:
        nvidia_smi = "nvidia-smi"

    # Get ID, processing and memory utilization for all GPUs
    try:
        p = Popen(
            [
                nvidia_smi,
                "--query-gpu=index,uuid,utilization.gpu,memory.total,"
                "memory.used,memory.free,driver_version,name,"
                "gpu_serial,display_active,display_mode,temperature.gpu",
                "--format=csv,noheader,nounits",
            ],
            stdout=PIPE,
        )
        stdout, stderror = p.communicate()
    except Exception:
        return []
    output = stdout.decode("UTF-8")
    # output = output[2:-1] # Remove b' and ' from string added by python
    # print(output)
    # Parse output
    # Split on line break
    lines = output.split(os.linesep)
    # print(lines)
    numDevices = len(lines) - 1
    GPUs = []
    for g in range(numDevices):
        line = lines[g]
        # print(line)
        vals = line.split(", ")
        # print(vals)
        for i in range(12):
            # print(vals[i])
            if i == 0:
                deviceIds = int(vals[i])
            elif i == 1:
                uuid = vals[i]
            elif i == 2:
                gpuUtil = safeFloatCast(vals[i]) / 100
            elif i == 3:
                memTotal = safeFloatCast(vals[i])
            elif i == 4:
                memUsed = safeFloatCast(vals[i])
            elif i == 5:
                memFree = safeFloatCast(vals[i])
            elif i == 6:
                driver = vals[i]
            elif i == 7:
                gpu_name = vals[i]
            elif i == 8:
                serial = vals[i]
            elif i == 9:
                display_active = vals[i]
            elif i == 10:
                display_mode = vals[i]
            elif i == 11:
                temp_gpu = safeFloatCast(vals[i])
        if all(math.isnan(mem) for mem in (memTotal, memUsed, memFree)):
            try:
                p = Popen(
                    [
                        nvidia_smi,
                        f"-i={deviceIds}"
                        f"--query-compute-apps=used_memory"
                        f"--format=csv,noheader,nounits",
                    ],
                    stdout=PIPE,
                )
                stdout, _ = p.communicate()
                output = stdout.decode("UTF-8")
                lines = output.split(os.linesep)
                mem = 0
                for line in lines:
                    if line:
                        mem += safeFloatCast(line)
                memUsed = mem
            except Exception:
                pass

        GPUs.append(
            GPU(
                deviceIds,
                uuid,
                gpuUtil,
                memTotal,
                memUsed,
                memFree,
                driver,
                gpu_name,
                serial,
                display_mode,
                display_active,
                temp_gpu,
            )
        )
    return GPUs  # (deviceIds, gpuUtil, memUtil)


def getAvailable(
        order="first",
        limit=1,
        maxLoad=0.5,
        maxMemory=0.5,
        memoryFree=0,
        includeNan=False,
        excludeID=[],
        excludeUUID=[],
):
    # order = first | last | random | load | memory first --> select the GPU
    # with the lowest ID (DEFAULT) last --> select the GPU with the highest
    # ID random --> select a random available GPU load --> select the GPU
    # with the lowest load memory --> select the GPU with the most memory
    # available limit = 1 (DEFAULT), 2, ..., Inf Limit sets the upper limit
    # for the number of GPUs to return. E.g. if limit = 2, but only one is
    # available, only one is returned.

    # Get device IDs, load and memory usage
    GPUs = getGPUs()

    # Determine, which GPUs are available
    GPUavailability = getAvailability(
        GPUs,
        maxLoad=maxLoad,
        maxMemory=maxMemory,
        memoryFree=memoryFree,
        includeNan=includeNan,
        excludeID=excludeID,
        excludeUUID=excludeUUID,
    )
    availAbleGPUindex = [
        idx
        for idx in range(0, len(GPUavailability))
        if (GPUavailability[idx] == 1)
    ]
    # Discard unavailable GPUs
    GPUs = [GPUs[g] for g in availAbleGPUindex]

    # Sort available GPUs according to the order argument
    if order == "first":
        GPUs.sort(
            key=lambda x: float("inf") if math.isnan(x.id) else x.id,
            reverse=False,
        )
    elif order == "last":
        GPUs.sort(
            key=lambda x: float("-inf") if math.isnan(x.id) else x.id,
            reverse=True,
        )
    elif order == "random":
        GPUs = [GPUs[g] for g in random.sample(range(0, len(GPUs)), len(GPUs))]
    elif order == "load":
        GPUs.sort(
            key=lambda x: float("inf") if math.isnan(x.load) else x.load,
            reverse=False,
        )
    elif order == "memory":
        GPUs.sort(
            key=lambda x: (
                float("inf") if math.isnan(x.memoryUtil) else x.memoryUtil
            ),
            reverse=False,
        )

    # Extract the number of desired GPUs, but limited to the total number of
    # available GPUs
    GPUs = GPUs[0: min(limit, len(GPUs))]

    # Extract the device IDs from the GPUs and return them
    deviceIds = [gpu.id for gpu in GPUs]

    return deviceIds


# def getAvailability(GPUs, maxLoad = 0.5, maxMemory = 0.5, includeNan =
# False): # Determine, which GPUs are available GPUavailability = np.zeros(
# len(GPUs)) for i in range(len(GPUs)): if (GPUs[i].load < maxLoad or (
# includeNan and np.isnan(GPUs[i].load))) and (GPUs[i].memoryUtil <
# maxMemory  or (includeNan and np.isnan(GPUs[i].memoryUtil))):
# GPUavailability[i] = 1


def getAvailability(
        GPUs,
        maxLoad=0.5,
        maxMemory=0.5,
        memoryFree=0,
        includeNan=False,
        excludeID=[],
        excludeUUID=[],
):
    # Determine, which GPUs are available
    GPUavailability = [
        (
            1
            if (gpu.memoryFree >= memoryFree) and (
                gpu.load < maxLoad or (
                    includeNan and math.isnan(gpu.load))) and (
                        gpu.memoryUtil < maxMemory or (
                            includeNan and math.isnan(gpu.memoryUtil))) and (
                                (gpu.id not in excludeID) and (
                                    gpu.uuid not in excludeUUID)
            )
            else 0
        )
        for gpu in GPUs
    ]
    return GPUavailability


def getFirstAvailable(
        order="first",
        maxLoad=0.5,
        maxMemory=0.5,
        attempts=1,
        interval=900,
        verbose=False,
        includeNan=False,
        excludeID=[],
        excludeUUID=[],
):
    for i in range(attempts):
        if verbose:
            print(
                "Attempting (" + str(i + 1) + "/" + str(attempts) + ") to "
                "locate available GPU."
            )
        # Get first available GPU
        available = getAvailable(
            order=order,
            limit=1,
            maxLoad=maxLoad,
            maxMemory=maxMemory,
            includeNan=includeNan,
            excludeID=excludeID,
            excludeUUID=excludeUUID,
        )
        # If an available GPU was found, break for loop.
        if available:
            if verbose:
                print("GPU " + str(available) + " located!")
            break
        # If this is not the last attempt, sleep for 'interval' seconds
        if i != attempts - 1:
            time.sleep(interval)
    # Check if an GPU was found, or if the attempts simply ran out. Throw
    # error, if no GPU was found
    if not (available):
        raise RuntimeError(
            "Could not find an available GPU after %s"
            " attempts with %s seconds interval." % (
                str(attempts),
                str(interval))
        )
    # Return found GPU
    return available


def showUtilization(all=False, attrList=None, useOldCode=False):
    GPUs = getGPUs()
    if all:
        if useOldCode:
            print(
                "ID | Name | Serial | UUID || GPU util. | Memory util. || "
                "Memory total | Memory used | Memory free || Display mode | "
                "Display active |"
            )
            print(
                "---------------------------------------------------------"
                "-------------------------------------------------------------"
                "--------"
            )
            for gpu in GPUs:
                print(
                    "{0:2d} | {1:s}  | {2:s} | {3:s} || "
                    "{4:3.0f}% | {5:3.0f}% || "
                    "{6:.0f}MB | {7:.0f}MB | {8:.0f}MB "
                    "|| {9:s}| {10:s}".format(
                        gpu.id,
                        gpu.name,
                        gpu.serial,
                        gpu.uuid,
                        gpu.load * 100,
                        gpu.memoryUtil * 100,
                        gpu.memoryTotal,
                        gpu.memoryUsed,
                        gpu.memoryFree,
                        gpu.display_mode,
                        gpu.display_active,
                    )
                )
        else:
            attrList = [
                [
                    {"attr": "id", "name": "ID"},
                    {"attr": "name", "name": "Name"},
                    {"attr": "serial", "name": "Serial"},
                    {"attr": "uuid", "name": "UUID"},
                ],
                [
                    {
                        "attr": "temperature",
                        "name": "GPU temp.",
                        "suffix": "C",
                        "transform": lambda x: x,
                        "precision": 0,
                    },
                    {
                        "attr": "load",
                        "name": "GPU util.",
                        "suffix": "%",
                        "transform": lambda x: x * 100,
                        "precision": 0,
                    },
                    {
                        "attr": "memoryUtil",
                        "name": "Memory util.",
                        "suffix": "%",
                        "transform": lambda x: x * 100,
                        "precision": 0,
                    },
                ],
                [
                    {
                        "attr": "memoryTotal",
                        "name": "Memory total",
                        "suffix": "MB",
                        "precision": 0,
                    },
                    {
                        "attr": "memoryUsed",
                        "name": "Memory used",
                        "suffix": "MB",
                        "precision": 0,
                    },
                    {
                        "attr": "memoryFree",
                        "name": "Memory free",
                        "suffix": "MB",
                        "precision": 0,
                    },
                ],
                [
                    {"attr": "display_mode", "name": "Display mode"},
                    {"attr": "display_active", "name": "Display active"},
                ],
            ]

    else:
        if useOldCode:
            print(" ID  GPU  MEM")
            print("--------------")
            for gpu in GPUs:
                print(
                    " {0:2d} {1:3.0f}% {2:3.0f}%".format(
                        gpu.id, gpu.load * 100, gpu.memoryUtil * 100
                    )
                )
        elif attrList is None:
            # if `attrList` was not specified, use the default one
            attrList = [
                [
                    {"attr": "id", "name": "ID"},
                    {
                        "attr": "load",
                        "name": "GPU",
                        "suffix": "%",
                        "transform": lambda x: x * 100,
                        "precision": 0,
                    },
                    {
                        "attr": "memoryUtil",
                        "name": "MEM",
                        "suffix": "%",
                        "transform": lambda x: x * 100,
                        "precision": 0,
                    },
                ],
            ]

    if not useOldCode:
        if attrList is not None:
            headerString = ""
            GPUstrings = [""] * len(GPUs)
            for attrGroup in attrList:
                # print(attrGroup)
                for attrDict in attrGroup:
                    headerString = headerString + "| " + attrDict["name"] + " "
                    headerWidth = len(attrDict["name"])
                    minWidth = len(attrDict["name"])

                    attrPrecision = (
                        "." + str(attrDict["precision"])
                        if ("precision" in attrDict.keys())
                        else ""
                    )
                    attrSuffix = (
                        str(attrDict["suffix"])
                        if ("suffix" in attrDict.keys())
                        else ""
                    )
                    attrTransform = (
                        attrDict["transform"]
                        if ("transform" in attrDict.keys())
                        else lambda x: x
                    )
                    for gpu in GPUs:
                        attr = getattr(gpu, attrDict["attr"])

                        attr = attrTransform(attr)

                        if isinstance(attr, float):
                            attrStr = ("{0:" + attrPrecision + "f}").format(
                                attr
                            )
                        elif isinstance(attr, int):
                            attrStr = ("{0:d}").format(attr)
                        elif isinstance(attr, str):
                            attrStr = attr
                        elif sys.version_info[0] == 2:
                            if isinstance(attr, str):
                                attrStr = attr.encode(
                                    "ascii", "ignore")
                        else:
                            raise TypeError(
                                "Unhandled object type (%s) "
                                "for attribute '%s'" % (
                                    str(type(attr)), attrDict["name"]
                                )
                            )

                        attrStr += attrSuffix
                        minWidth = max(minWidth, len(attrStr))
                    headerString += " " * max(0, minWidth - headerWidth)
                    minWidthStr = str(minWidth - len(attrSuffix))

                    for gpuIdx, gpu in enumerate(GPUs):
                        attr = getattr(gpu, attrDict["attr"])

                        attr = attrTransform(attr)

                        if isinstance(attr, float):
                            attrStr = (
                                "{0:" + minWidthStr + attrPrecision + "f}"
                            ).format(attr)
                        elif isinstance(attr, int):
                            attrStr = ("{0:" + minWidthStr + "d}").format(attr)
                        elif isinstance(attr, str):
                            attrStr = ("{0:" + minWidthStr + "s}").format(attr)
                        elif sys.version_info[0] == 2:
                            if isinstance(attr, str):
                                attrStr = ("{0:" + minWidthStr + "s}").format(
                                    attr.encode(
                                        "ascii", "ignore")
                                )
                        else:
                            raise TypeError(
                                "Unhandled object type (%s) "
                                "for attribute '%s'" % (
                                    str(type(attr)),
                                    attrDict["name"]
                                )
                            )

                        attrStr += attrSuffix

                        GPUstrings[gpuIdx] += "| " + attrStr + " "

                headerString = headerString + "|"
                for gpuIdx, gpu in enumerate(GPUs):
                    GPUstrings[gpuIdx] += "|"

            headerSpacingString = "-" * len(headerString)
            print(headerString)
            print(headerSpacingString)
            for GPUstring in GPUstrings:
                print(GPUstring)



import concurrent.futures
import os
import sys
import subprocess

from typing import Optional

from kiroframe_arcee.utils import run_async

EXECUTABLE_DIR = os.path.dirname(os.path.realpath(sys.argv[0]))


class Collector:
    executor = concurrent.futures.ThreadPoolExecutor(max_workers=3)

    @staticmethod
    def __get_remote() -> str:
        return subprocess.check_output(
            ["git", "config", "--get", "remote.origin.url"],
            cwd=EXECUTABLE_DIR
        ).decode('UTF-8').strip()

    @staticmethod
    def __get_branch() -> str:
        return subprocess.check_output(
            ["git", "branch", "--show-current"],
            cwd=EXECUTABLE_DIR).decode('UTF-8').strip()

    @staticmethod
    def __get_commit_id() -> str:
        return subprocess.check_output(
            ["git", "rev-parse", "HEAD"],
            cwd=EXECUTABLE_DIR).decode('UTF-8').strip()

    @staticmethod
    def __get_status() -> str:
        try:
            subprocess.check_call(["git", "diff", "--exit-code", "--quiet"],
                                  cwd=EXECUTABLE_DIR)
            return "clean"
        except subprocess.CalledProcessError:
            return "dirty"

    @classmethod
    def _collect(cls) -> Optional[dict]:
        try:
            return {
                "remote": cls.__get_remote(),
                "branch": cls.__get_branch(),
                "commit_id": cls.__get_commit_id(),
                "status": cls.__get_status()
            }
        except (FileNotFoundError, subprocess.CalledProcessError):
            return

    @classmethod
    async def collect(cls):
        return await run_async(cls._collect, executor=cls.executor)



import os
import asyncio
import threading
from typing import List, Dict
from kiroframe_arcee.modules.providers import local_file, amazon

LOCAL_PREFIX = 'file://'
S3_PREFIX = 's3://'
BASE_PATH = 'kiroframe/datasets/%s/'


class DatasetThread(threading.Thread):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.exception = None

    def run(self):
        try:
            if self._target:
                self._target(*self._args, **self._kwargs)
        except Exception as e:
            self.exception = e


class Dataset(object):
    __slots__ = ('key', 'name', 'description', 'labels', 'meta',
                 'timespan_from', 'timespan_to', 'aliases',
                 '_tasks', '_files', '_version')

    def __init__(self, key: str, name: str = None, description: str = None,
                 labels: List[str] = None, meta: Dict = None,
                 timespan_from: int = None, timespan_to: int = None,
                 aliases: List[str] = None):
        self._tasks: List = []
        self._files: Dict = {}
        self._version: int = None
        self.key: str = key
        self.name: str = name
        self.description: str = description
        self.labels: List[str] = labels or []
        self.meta: Dict = meta or {}
        self.timespan_from: int = timespan_from
        self.timespan_to: int = timespan_to
        self.aliases: List[str] = aliases or []

    @classmethod
    def from_response(cls, response):
        version = response.pop('version', {})
        files = version.pop('files', [])
        response.update(version)
        obj = cls(**{
            k: response.get(k) for k in cls.__slots__ if k in response
        })
        obj._version = version['version']
        if files:
            obj._files = {f['path']: {
                'path': f['path'],
                'size': f['size'],
                'digest': f['digest']
            } for f in files}
        return obj

    @property
    def __dict__(self):
        res = dict()
        for k in self.__slots__:
            if k.startswith('_'):
                continue
            value = getattr(self, k)
            if value:
                res[k] = getattr(self, k)
        res['files'] = list(self._files.values())
        return res

    def _get_provider(self, path):
        if path.startswith(LOCAL_PREFIX):
            return local_file, path.strip(LOCAL_PREFIX)
        elif path.startswith(S3_PREFIX):
            return amazon, path
        else:
            raise TypeError('Unhandled path type')

    def _add_file(self, path):
        provider, local_path = self._get_provider(path)
        digest, size = asyncio.run(provider.get_file_info(local_path))
        self._files[path] = {
            'path': path,
            'size': size,
            'digest': digest
        }

    def add_file(self, path):
        if path in self._files and self._version is None:
            return
        self._version = None
        self._files[path] = None
        thr = DatasetThread(target=self._add_file, args=(path, ))
        thr.start()
        self._tasks.append(thr)

    def remove_file(self, path):
        if path in self._files:
            del self._files[path]

    def wait_ready(self):
        for task in self._tasks:
            task.join()
            if task.exception:
                raise task.exception

    def download(self, overwrite=True) -> dict:
        download_map = dict()
        if self._version is None:
            raise TypeError('Dataset is not logged')
        name = f'{self.key}:V{self._version}'
        print('Downloading %s' % name)
        for path, file in self._files.items():
            destination = BASE_PATH % name
            file_name = path.split('/')[-1]
            download_path = destination + file_name
            download_map[path] = download_path
            if not overwrite and os.path.isfile(download_path):
                continue
            thr = DatasetThread(
                target=self._download, args=(file, destination, file_name))
            thr.start()
            self._tasks.append(thr)
        self.wait_ready()
        print('Download completed: %s' % name)
        return download_map

    def _download(self, file, destination, file_name):
        digest = file['digest']
        path = file['path']
        provider, local_path = self._get_provider(path)
        asyncio.run(
            provider.download(
                local_path, digest, destination, file_name
            )
        )



from aiounittest import AsyncTestCase
from unittest.mock import patch

from kiroframe_arcee.platform import (
    Platform,
    AzureCollector,
    CollectorFactory,
    PlatformType,
    AlibabaCollector,
    AwsCollector,
    GcpCollector,
    PlatformMeta,
    InstanceLifeCycle,
)
from tests.test_data import TestDataAzure


class TestPlatform(AsyncTestCase):
    def test_platform_is_static(self):
        self.assertRaises(TypeError, Platform())

    @patch("kiroframe_arcee.platform.Platform.board_version")
    async def test_platform_factory(self, mock_board_version):
        mock_board_version.return_value = PlatformType.aws
        self.assertEqual(await CollectorFactory.get(), AwsCollector)


class TestAzureCollector(AsyncTestCase):
    @patch("kiroframe_arcee.platform.AzureCollector.collect")
    async def test_azure_base_collector(self, mocked_collect):
        mocked_collect.return_value = TestDataAzure.get_test_data()
        platform_meta = await AzureCollector().get_platform_meta()
        self.assertTrue(isinstance(platform_meta, PlatformMeta))
        self.assertTrue(platform_meta.platform_type, PlatformType.azure)
        self.assertTrue(platform_meta.local_ip)
        self.assertTrue(platform_meta.instance_lc, InstanceLifeCycle.Unknown)
        self.assertTrue(platform_meta.to_dict())

    @patch("kiroframe_arcee.platform.AzureCollector.collect")
    async def test_azure_no_network(self, mocked_collect):
        mocked_collect.return_value = TestDataAzure.get_test_data_no_network()
        platform_meta = await AzureCollector().get_platform_meta()
        self.assertTrue(isinstance(platform_meta, PlatformMeta))
        self.assertTrue(platform_meta.platform_type, PlatformType.azure)
        self.assertFalse(platform_meta.public_ip)
        self.assertFalse(platform_meta.local_ip)
        self.assertTrue(platform_meta.instance_lc, InstanceLifeCycle.Unknown)
        self.assertTrue(platform_meta.to_dict())

    @patch("kiroframe_arcee.platform.AzureCollector.collect")
    async def test_azure_no_data(self, mocked_collect):
        mocked_collect.return_value = None
        platform_meta = await AzureCollector().get_platform_meta()
        self.assertTrue(isinstance(platform_meta, PlatformMeta))
        self.assertTrue(platform_meta.platform_type, PlatformType.azure)
        self.assertTrue(platform_meta.instance_lc, InstanceLifeCycle.Unknown)
        self.assertTrue(platform_meta.to_dict())


class TestAwsCollector(AsyncTestCase):
    @patch("kiroframe_arcee.platform.AwsCollector.get_instance_type")
    @patch("kiroframe_arcee.platform.AwsCollector.get_region")
    @patch("kiroframe_arcee.platform.AwsCollector.get_az")
    @patch("kiroframe_arcee.platform.AwsCollector.get_life_cycle")
    @patch("kiroframe_arcee.platform.AwsCollector.get_public_ip")
    @patch("kiroframe_arcee.platform.AwsCollector.get_local_ip")
    @patch("kiroframe_arcee.platform.AwsCollector.get_account_id")
    @patch("kiroframe_arcee.platform.AwsCollector.get_instance_id")
    async def test_aws_collector_iface(
        self,
        m_instance_id,
        m_account_id,
        m_local_ip,
        m_public_ip,
        m_life_cycle,
        m_az,
        m_region,
        m_type,
    ):
        m_instance_id.return_value = "i-09dc9f5553f84a9ad"
        m_account_id.return_value = "00000000000"
        m_local_ip.return_value = "172.31.24.6"
        m_public_ip.return_value = "1.1.1.1"
        m_life_cycle.return_value = InstanceLifeCycle.OnDemand
        m_az.return_value = "eu-central-1a"
        m_region.return_value = "eu-central-1a"
        m_type.return_value = "m6in.large"
        platform_meta = await AwsCollector().get_platform_meta()
        self.assertTrue(platform_meta.platform_type, PlatformType.aws)
        self.assertTrue(platform_meta.instance_lc, InstanceLifeCycle.OnDemand)
        self.assertTrue(platform_meta.to_dict())


class TestAlibabaCollector(AsyncTestCase):
    @patch("kiroframe_arcee.platform.AlibabaCollector.get_instance_type")
    @patch("kiroframe_arcee.platform.AlibabaCollector.get_region")
    @patch("kiroframe_arcee.platform.AlibabaCollector.get_az")
    @patch("kiroframe_arcee.platform.AlibabaCollector.get_life_cycle")
    @patch("kiroframe_arcee.platform.AlibabaCollector.get_public_ip")
    @patch("kiroframe_arcee.platform.AlibabaCollector.get_local_ip")
    @patch("kiroframe_arcee.platform.AlibabaCollector.get_account_id")
    @patch("kiroframe_arcee.platform.AlibabaCollector.get_instance_id")
    async def test_ali_collector_iface(
        self,
        m_instance_id,
        m_account_id,
        m_local_ip,
        m_public_ip,
        m_life_cycle,
        m_az,
        m_region,
        m_type,
    ):
        m_instance_id.return_value = "i-gw8csaxjubfr17s2e1ip"
        m_account_id.return_value = "00000000000"
        m_local_ip.return_value = "192.168.0.83"
        m_public_ip.return_value = "2.2.2.2"
        m_life_cycle.return_value = InstanceLifeCycle.Spot
        m_az.return_value = "eu-central-1a"
        m_region.return_value = "eu-central-1"
        m_type.return_value = "ecs.t5-c1m1.large"
        platform_meta = await AlibabaCollector().get_platform_meta()
        self.assertTrue(platform_meta.platform_type, PlatformType.alibaba)
        self.assertTrue(platform_meta.instance_lc, InstanceLifeCycle.Spot)
        self.assertTrue(platform_meta.to_dict())


class TestGcpCollector(AsyncTestCase):
    @patch("kiroframe_arcee.platform.GcpCollector.get_instance_type")
    @patch("kiroframe_arcee.platform.GcpCollector.get_locations")
    @patch("kiroframe_arcee.platform.GcpCollector.get_life_cycle")
    @patch("kiroframe_arcee.platform.GcpCollector.get_public_ip")
    @patch("kiroframe_arcee.platform.GcpCollector.get_local_ip")
    @patch("kiroframe_arcee.platform.GcpCollector.get_account_id")
    @patch("kiroframe_arcee.platform.GcpCollector.get_instance_id")
    async def test_gcp_collector_iface(
        self,
        m_instance_id,
        m_account_id,
        m_local_ip,
        m_public_ip,
        m_life_cycle,
        m_locations,
        m_type,
    ):
        m_instance_id.return_value = "3358770582979744445"
        m_account_id.return_value = "accountid"
        m_local_ip.return_value = "10.194.0.5"
        m_public_ip.return_value = "3.3.3.3"
        m_life_cycle.return_value = InstanceLifeCycle.OnDemand
        m_locations.return_value = ("southamerica-west1",
                                    "southamerica-west1-a")
        m_type.return_value = "e2-standard-4"
        platform_meta = await GcpCollector().get_platform_meta()
        self.assertTrue(platform_meta.platform_type, PlatformType.gcp)
        self.assertTrue(platform_meta.instance_lc, InstanceLifeCycle.OnDemand)
        self.assertTrue(platform_meta.to_dict())



import os
import sys
import hashlib
import shutil
import mmap
import aiofiles

_KB: int = 1_024
_CHUNKSIZE: int = 128 * _KB


async def get_file_info(path):
    digest = await _get_md5(path)
    size = await _get_size(path)
    return digest, size


async def _get_md5(path):
    if sys.version_info >= (3, 9):
        md_5_hash = hashlib.md5(usedforsecurity=False)
    else:
        md_5_hash = hashlib.md5()
    async with aiofiles.open(path, "rb") as f:
        try:
            with mmap.mmap(f.fileno(), length=0,
                           access=mmap.ACCESS_READ) as mview:
                md_5_hash.update(mview)
        except OSError:
            chunk = f.read(_CHUNKSIZE)
            while chunk:
                md_5_hash.update(chunk)
                chunk = f.read(_CHUNKSIZE)
        except ValueError:
            pass
    return md_5_hash.hexdigest()


async def _get_size(path):
    st = os.stat(path)
    return st.st_size


async def download(path, digest, dest_path, file_name):
    if not os.path.exists(path):
        raise ValueError('Failed to find file path %s' % path)
    md5 = await _get_md5(path)
    if md5 != digest:
        raise ValueError(
            'Cannot download dataset file %s. Source file has been changed' %
            path)
    os.makedirs(dest_path, exist_ok=True)
    shutil.copy(path, dest_path + file_name)



import os
import aioboto3
import aiofiles
from urllib.parse import urlparse


async def get_file_info(path):
    bucket, key = await _parse_uri(path)
    session = aioboto3.Session()
    async with session.resource("s3") as s3:
        res = await s3.Object(bucket, key)
        etag = await res.e_tag
        size = await res.content_length
    return etag.strip('"'), size


async def _parse_uri(path):
    url = urlparse(path)
    bucket = url.netloc
    key = url.path[1:]
    return bucket, key


async def download(path, digest, dest_path, file_name):
    bucket, key = await _parse_uri(path)
    session = aioboto3.Session()
    async with session.resource("s3") as s3:
        s3_object = await s3.Object(bucket, key)
        etag = await s3_object.e_tag
        if etag.strip('"') != digest:
            raise ValueError(
                'Cannot download dataset file %s. Source file has been '
                'changed' % path)
        os.makedirs(dest_path, exist_ok=True)
        async with aiofiles.open(dest_path + file_name, 'wb') as fp:
            await s3_object.download_fileobj(fp)


async def main(bucket, key):
    session = aioboto3.Session()
    async with session.resource("s3") as s3:
        bucket = await s3.Object(bucket, key)
        async with aiofiles.open('TEST', 'wb') as fp:
            await bucket.download_fileobj(fp)



# flake8: noqa: F401
from .arcee import (init, send, tag, milestone, info, finish, error, stage,
                    hyperparam, model, model_version, model_version_alias,
                    model_version_tag, artifact, artifact_tag, Dataset,
                    log_dataset, use_dataset)



import asyncio
from concurrent.futures import ThreadPoolExecutor
from functools import partial


def single(class_):
    instances = {}

    def get_instance(*args, **kwargs):
        if class_ not in instances:
            instances[class_] = class_(*args, **kwargs)
        return instances[class_]

    return get_instance


async def run_async(func, *args, loop=None, executor=None, **kwargs):
    if loop is None:
        loop = asyncio.get_event_loop()
    if executor is None:
        executor = ThreadPoolExecutor(max_workers=10)
    pfunc = partial(func, *args, **kwargs)
    return await loop.run_in_executor(executor, pfunc)



import concurrent.futures
from io import StringIO
import sys
from typing import Dict

from kiroframe_arcee.utils import run_async


class StdProxy:
    def __init__(self, std_stream, on_op_cbs):
        self._std_stream = std_stream
        self._on_op_cbs = on_op_cbs

    def __getattr__(self, name):
        return self._wrapped(name)

    def _wrapped(self, name):
        cb_op = self._on_op_cbs.get(name)
        if not cb_op:
            return getattr(self._std_stream, name)

        def inner(*args, **kwargs):
            cb_op(*args, **kwargs)
            return getattr(self._std_stream, name)(*args, **kwargs)

        return inner


class WritesCollector:
    def __init__(self, std_stream):
        self.stream = StringIO()
        self.cb_map = {
            'write': self.handle_write()
        }
        self.proxy = StdProxy(std_stream, self.cb_map)

    def handle_write(self):
        def inner(*args, **kwargs):
            for arg in args:
                # print typically writes args
                self.stream.write(arg)

        return inner

    def get_writes(self):
        return self.stream.getvalue()


stdout_writes = WritesCollector(sys.stdout)
stderr_writes = WritesCollector(sys.stderr)


def acquire_console():
    sys.stdout = stdout_writes.proxy
    sys.stderr = stderr_writes.proxy


def release_console():
    sys.stdout = sys.__stdout__
    sys.stderr = sys.__stderr__


class Collector:
    executor = concurrent.futures.ThreadPoolExecutor(max_workers=3)

    @classmethod
    def _collect(cls) -> Dict[str, str]:
        return {
            "output": stdout_writes.get_writes(),
            "error": stderr_writes.get_writes()
        }

    @classmethod
    async def collect(cls):
        return await run_async(cls._collect, executor=cls.executor)



import GPUtil as GPU
import sys

# Get all device ids and their processing and memory utiliazion
# (deviceIds, gpuUtil, memUtil) = GPU.getGPUs()

# Print os and python version information
print("OS: " + sys.platform)
print(sys.version)

# Print package name and version number
print(GPU.__name__ + " " + GPU.__version__)

# Show the utilization of all GPUs in a nice table
GPU.showUtilization()

# Show all stats of all GPUs in a nice table
GPU.showUtilization(all=True)

# Get all available GPU(s), ordered by ID in ascending order
print("All available ordered by id: "),
print(GPU.getAvailable(order="first", limit=999))

# Get 1 available GPU, ordered by ID in descending order
print("Last available: "),
print(GPU.getAvailable(order="last", limit=1))

# Get 1 random available GPU
print("Random available: "),
print(GPU.getAvailable(order="random"))

# Get 1 available GPU, ordered by GPU load ascending
print("First available weighted by GPU load ascending: "),
print(GPU.getAvailable(order="load", limit=1))

# Get all available GPU with max load of 10%, ordered by memory ascending
print("All available weighted by memory load ascending: "),
print(GPU.getAvailable(order="memory", limit=999, maxLoad=0.1))

# Get the first available GPU
firstGPU = GPU.getFirstAvailable()
print("First available GPU id:"),
print(firstGPU)

# Get the first available GPU, where memory usage is less than 90% and
# processing is less than 80%
firstGPU = GPU.getFirstAvailable(maxMemory=0.9, maxLoad=0.8)
print("First available GPU id (memory < 90%, load < 80%):"),
print(firstGPU)

# Get the first available GPU, where processing is less than 1%
firstGPU = GPU.getFirstAvailable(
    attempts=5, interval=5, maxLoad=0.01, verbose=True
)
print("First available GPU id (load < 1%):"),
print(firstGPU)
# NOTE: If all your GPUs currently have a load larger than 1%, this step
# will fail. It's not a bug! It is intended to do so, if it does not find an
# available GPU.

# Get the first available GPU, where memory usage is less than 1%
firstGPU = GPU.getFirstAvailable(
    attempts=5, interval=5, maxMemory=0.01, verbose=True
)
print("First available GPU id (memory < 1%):"),
print(firstGPU)
# NOTE: If all your GPUs currently have a memory consumption larger than 1%,
# this step will fail. It's not a bug! It is intended to do so, if it does not
# find an available GPU.
